from pathlib import Path

import yaml
from truss.local.local_config_handler import LocalConfigHandler
from truss.patch.hash import str_hash_str
from truss.truss_handle import TrussHandle
from truss.util.path import copy_file_path, copy_tree_path, remove_tree_path


def gather(truss_path: Path) -> Path:
    handle = TrussHandle(truss_path)
    shadow_truss_dir_name = calc_shadow_truss_dirname(truss_path)
    shadow_truss_metdata_file_path = (
        LocalConfigHandler.shadow_trusses_dir_path()
        / f"{shadow_truss_dir_name}.metadata.yaml"
    )
    shadow_truss_path = (
        LocalConfigHandler.shadow_trusses_dir_path() / shadow_truss_dir_name
    )
    if shadow_truss_metdata_file_path.exists():
        with shadow_truss_metdata_file_path.open() as fp:
            metadata = yaml.safe_load(fp)
        max_mod_time = metadata["max_mod_time"]
        if max_mod_time == handle.max_modified_time:
            return shadow_truss_path

        # Shadow truss is out of sync, clear it
        shadow_truss_metdata_file_path.unlink()
        remove_tree_path(shadow_truss_path)

    copy_tree_path(truss_path, shadow_truss_path)
    packages_dir_path_in_shadow = (
        shadow_truss_path / handle.spec.config.bundled_packages_dir
    )
    packages_dir_path_in_shadow.mkdir(exist_ok=True)
    for path in handle.spec.external_package_dirs_paths:
        if not path.is_dir():
            raise ValueError(
                f"External packages directory at {path} is not a directory"
            )
        # We copy over contents of the external package directory, not the
        # directory itself. This mimics the local load behavior and is meant to
        # replicate adding external package directory to sys.path which doesn't
        # make the directory available as a package to python but the contents
        # inside.
        #
        # Note that this operation can fail if there are conflicts. Onus is on
        # the creator of truss to make sure that there are no conflicts.
        for sub_path in path.iterdir():
            if sub_path.is_dir():
                copy_tree_path(sub_path, packages_dir_path_in_shadow / sub_path.name)
            if sub_path.is_file():
                copy_file_path(sub_path, packages_dir_path_in_shadow / sub_path.name)

    # Don't run validation because they will fail until we clear external
    # packages. We do it after.
    shadow_handle = TrussHandle(shadow_truss_path, validate=False)
    shadow_handle.clear_external_packages()
    shadow_handle.validate()
    with shadow_truss_metdata_file_path.open("w") as fp:
        yaml.safe_dump({"max_mod_time": handle.max_modified_time}, fp)
    return shadow_truss_path


def calc_shadow_truss_dirname(truss_path: Path) -> str:
    resolved_path_str = str(truss_path.resolve())
    return str_hash_str(resolved_path_str)



import random
import string
from pathlib import Path
from typing import Callable, List

import pytest
from truss.patch.hash import (
    directory_content_hash,
    file_content_hash,
    file_content_hash_str,
)


@pytest.fixture
def dir_hash_test_dir(tmp_path: Path):
    target_dir = tmp_path / "target"
    target_dir.mkdir()
    target_dir_file = target_dir / "target_file"
    _update_file_content(target_dir_file)
    sub_dir = target_dir / "subdir"
    sub_dir.mkdir()
    sub_dir_file = sub_dir / "subdir_file"
    _update_file_content(sub_dir_file)
    return target_dir


def test_dir_hash_same_on_retry(dir_hash_test_dir):
    _verify_with_dir_modification(dir_hash_test_dir, lambda _: None, True)


def test_dir_hash_different_if_file_added_at_top_level(dir_hash_test_dir):
    def mod(dir_path):
        _update_file_content(dir_path / "new_file")

    _verify_with_dir_modification(dir_hash_test_dir, mod, False)


def test_dir_hash_different_if_file_added_to_sub_directory(dir_hash_test_dir):
    def mod(dir_path):
        _update_file_content(dir_path / "subdir" / "new_file")

    _verify_with_dir_modification(dir_hash_test_dir, mod, False)


def test_dir_hash_different_if_file_file_content_modified(dir_hash_test_dir):
    def mod(dir_path):
        _update_file_content(
            dir_path / "target_file", content=_generate_random_string(5)
        )

    _verify_with_dir_modification(dir_hash_test_dir, mod, False)


def test_dir_hash_same_if_target_dir_renamed(tmp_path, dir_hash_test_dir):
    def mod(dir_path):
        target = tmp_path / "renamed"
        dir_path.rename(target)
        return target

    _verify_with_dir_modification(dir_hash_test_dir, mod, True)


def test_dir_hash_same_if_target_dir_moved_but_not_renamed(tmp_path, dir_hash_test_dir):
    def mod(dir_path):
        holding_dir = tmp_path / "holding_dir"
        holding_dir.mkdir()
        target = holding_dir / dir_path.name
        dir_path.rename(target)
        return target

    _verify_with_dir_modification(dir_hash_test_dir, mod, True)


def test_dir_hash_different_if_sub_directory_renamed(dir_hash_test_dir):
    def mod(dir_path):
        sub_dir = dir_path / "subdir"
        sub_dir.rename(dir_path / "new_subdir")

    _verify_with_dir_modification(dir_hash_test_dir, mod, False)


def test_dir_hash_different_if_file_contents_swapped(dir_hash_test_dir):
    _update_file_content(dir_hash_test_dir / "file1")
    _update_file_content(dir_hash_test_dir / "file2")

    def mod(dir_path):
        tmp_file = dir_path / "tmp_file"
        (dir_path / "file1").rename(tmp_file)
        (dir_path / "file2").rename(dir_path / "file1")
        tmp_file.rename(dir_path / "file2")

    _verify_with_dir_modification(dir_hash_test_dir, mod, False)


def test_dir_hash_ignore_pattern_filename(dir_hash_test_dir):
    _update_file_content(dir_hash_test_dir / "file1")
    _update_file_content(dir_hash_test_dir / "file2")

    def mod_file1(dir_path):
        _update_file_content(dir_path / "file1")

    def mod_file2(dir_path):
        _update_file_content(dir_path / "file2")

    _verify_with_dir_modification(dir_hash_test_dir, mod_file1, True, ["file1"])
    _verify_with_dir_modification(dir_hash_test_dir, mod_file2, False, ["file1"])


def test_dir_hash_ignore_pattern_multiple(dir_hash_test_dir):
    _update_file_content(dir_hash_test_dir / "file1")
    _update_file_content(dir_hash_test_dir / "file2")

    def mod(dir_path):
        _update_file_content(dir_path / "file1")
        _update_file_content(dir_path / "file2")

    _verify_with_dir_modification(dir_hash_test_dir, mod, True, ["file1", "file2"])


def test_dir_hash_ignore_pattern_dir_glob(dir_hash_test_dir):
    tmp_dir = dir_hash_test_dir / "tmp_dir"
    tmp_dir.mkdir()
    _update_file_content(tmp_dir / "file1")
    _update_file_content(tmp_dir / "file2")

    def mod(dir_path):
        _update_file_content(dir_path / "tmp_dir" / "file1")
        _update_file_content(dir_path / "tmp_dir" / "file2")

    _verify_with_dir_modification(dir_hash_test_dir, mod, True, ["tmp_dir/*"])


def test_file_content_hash(tmp_path):
    orig_content = _generate_random_string(1024 * 1024)
    file_path = tmp_path / "file"
    _update_file_content(file_path, orig_content)
    orig_hash = file_content_hash(file_path)
    new_content = _generate_random_string(2048 * 1024)
    _update_file_content(file_path, new_content)
    new_hash = file_content_hash(file_path)
    assert orig_hash != new_hash

    _update_file_content(file_path, orig_content)
    final_hash = file_content_hash(file_path)
    assert final_hash == orig_hash


def test_file_content_hash_str(tmp_path):
    orig_content = _generate_random_string(1024 * 1024)
    file_path = tmp_path / "file"
    _update_file_content(file_path, orig_content)
    orig_hash = file_content_hash_str(file_path)
    new_content = _generate_random_string(2048 * 1024)
    _update_file_content(file_path, new_content)
    new_hash = file_content_hash_str(file_path)
    assert orig_hash != new_hash

    _update_file_content(file_path, orig_content)
    final_hash = file_content_hash_str(file_path)
    assert final_hash == orig_hash


def _verify_with_dir_modification(
    target_dir: Path,
    op: Callable[[Path], Path],
    should_match: bool,
    ignore_patterns: List[str] = None,
):
    hash1 = directory_content_hash(target_dir, ignore_patterns=ignore_patterns)
    new_target_dir = op(target_dir)
    hash2 = directory_content_hash(
        new_target_dir or target_dir, ignore_patterns=ignore_patterns
    )
    if should_match:
        assert hash1 == hash2
    else:
        assert hash1 != hash2


def _generate_random_string(size: int = 10) -> str:
    letters = string.ascii_letters
    return "".join(random.choice(letters) for _ in range(size))


def _update_file_content(file: Path, content: str = None):
    if content is None:
        content = _generate_random_string()
    with file.open("w") as f:
        f.write(content)



import warnings

# These warnings show up if we have pydantic v2 installed, but still use v1-compatible
# APIs, which are of course deprecated...
warnings.filterwarnings("ignore", message="Valid config keys have changed in V2:*")
warnings.filterwarnings(
    "ignore",
    message="`pydantic.generics:GenericModel` has been moved to `pydantic.BaseModel`.",
)

del warnings

# flake8: noqa F401
from truss_chains.definitions import (
    Assets,
    ChainsRuntimeError,
    Compute,
    DeploymentContext,
    DockerImage,
    RemoteConfig,
    RemoteErrorDetail,
    RPCOptions,
)
from truss_chains.public_api import (
    ChainletBase,
    depends,
    depends_context,
    deploy_remotely,
    entrypoint,
    run_local,
)
from truss_chains.stub import StubBase
from truss_chains.utils import make_abs_path_here



import tarfile
import tempfile
from pathlib import Path
from typing import IO, Any, Callable, List

from rich.progress import Progress
from truss.util.path import is_ignored


class ReadProgressIndicatorFileHandle:
    def __init__(
        self, file: IO[bytes], progress_callback: Callable[[int], Any]
    ) -> None:
        self._file = file
        self._progress_callback = progress_callback

    def read(self, size=-1):
        data = self._file.read(size)
        self._progress_callback(len(data))
        return data

    def __getattr__(self, attr):
        return getattr(self._file, attr)


def create_tar_with_progress_bar(
    source_dir: Path, ignore_patterns: List[str] = [], delete=True
):
    # Exclude files that match the ignore_patterns
    files_to_include = [
        f
        for f in source_dir.rglob("*")
        if f.is_file() and not is_ignored(f, ignore_patterns, source_dir)
    ]

    total_size = sum(f.stat().st_size for f in files_to_include)
    temp_file = tempfile.NamedTemporaryFile(suffix=".tgz", delete=delete)

    with tarfile.open(temp_file.name, "w:") as tar:

        progress = Progress()

        task_id = progress.add_task("[cyan]Compressing...", total=total_size)

        with progress:

            def file_read_progress_callback(bytes_read: int):
                progress.update(task_id, advance=bytes_read)

            for file_path in files_to_include:
                arcname = str(file_path.relative_to(source_dir))
                with file_path.open("rb") as file_obj:
                    file_obj_with_progress = ReadProgressIndicatorFileHandle(
                        file_obj, file_read_progress_callback
                    )
                    tarinfo = tar.gettarinfo(name=str(file_path), arcname=arcname)
                    tar.addfile(
                        tarinfo=tarinfo, fileobj=file_obj_with_progress  # type: ignore[arg-type]
                    )
    return temp_file



import pytest
from truss.templates.control.control.helpers.truss_patch.requirement_name_identifier import (
    identify_requirement_name,
    reqs_by_name,
)


@pytest.mark.parametrize(
    "req, expected_name",
    [
        ("pytorch", "pytorch"),
        (
            "git+https://github.com/huggingface/transformers.git",
            "git+https://github.com/huggingface/transformers.git",
        ),
        (
            " git+https://github.com/huggingface/transformers.git ",
            "git+https://github.com/huggingface/transformers.git",
        ),
        ("pytorch==1.0", "pytorch"),
        ("pytorch>=1.0", "pytorch"),
        ("pytorch<=1.0", "pytorch"),
    ],
)
def test_identify_requirement_name(req, expected_name):
    assert expected_name == identify_requirement_name(req)


def test_reqs_by_name():
    reqs = [
        "pytorch",
        "jinja==1.0",
    ]
    assert reqs_by_name(reqs) == {"pytorch": "pytorch", "jinja": "jinja==1.0"}



from typing import Optional

import numpy as np
import tritonclient
import tritonclient.grpc.aio as grpcclient


class ModelInput:
    def __init__(
        self,
        prompt: str,
        request_id: int,
        max_tokens: int = 50,
        max_new_tokens: int = 50,
        temperature: float = 1.0,
        top_p: float = 1.0,
        top_k: int = 50,
        beam_width: int = 1,
        bad_words_list: Optional[list] = None,
        stop_words_list: Optional[list] = None,
        repetition_penalty: float = 1.0,
        ignore_eos: bool = False,
        stream: bool = True,
        eos_token_id: int = None,  # type: ignore
    ) -> None:
        self.stream = stream
        self.request_id = request_id
        self._prompt = prompt
        self._max_tokens = max_tokens
        self._beam_width = beam_width
        self._bad_words_list = [""] if bad_words_list is None else bad_words_list
        self._stop_words_list = [""] if stop_words_list is None else stop_words_list
        self._repetition_penalty = repetition_penalty
        self._eos_token_id = eos_token_id
        self._ignore_eos = ignore_eos
        # These variables are passed by OAI proxy but are unused
        # TODO(Abu): Add support for these
        self._max_new_tokens = max_new_tokens
        self._temperature = temperature
        self._top_p = top_p
        self._top_k = top_k

    def _prepare_grpc_tensor(
        self, name: str, input_data: np.ndarray
    ) -> grpcclient.InferInput:
        tensor = grpcclient.InferInput(
            name,
            input_data.shape,
            tritonclient.utils.np_to_triton_dtype(input_data.dtype),
        )
        tensor.set_data_from_numpy(input_data)
        return tensor

    def to_tensors(self):
        if self._eos_token_id is None and self._ignore_eos:
            raise ValueError("eos_token_id is required when ignore_eos is True")

        prompt_data = np.array([[self._prompt]], dtype=object)
        output_len_data = np.ones_like(prompt_data, dtype=np.int32) * self._max_tokens
        bad_words_data = np.array([self._bad_words_list], dtype=object)
        stop_words_data = np.array([self._stop_words_list], dtype=object)
        stream_data = np.array([[self.stream]], dtype=bool)
        beam_width_data = np.array([[self._beam_width]], dtype=np.int32)
        repetition_penalty_data = np.array(
            [[self._repetition_penalty]], dtype=np.float32
        )

        inputs = [
            self._prepare_grpc_tensor("text_input", prompt_data),
            self._prepare_grpc_tensor("max_tokens", output_len_data),
            self._prepare_grpc_tensor("bad_words", bad_words_data),
            self._prepare_grpc_tensor("stop_words", stop_words_data),
            self._prepare_grpc_tensor("stream", stream_data),
            self._prepare_grpc_tensor("beam_width", beam_width_data),
            self._prepare_grpc_tensor("repetition_penalty", repetition_penalty_data),
        ]

        if not self._ignore_eos:
            end_id_data = np.array([[self._eos_token_id]], dtype=np.int32)
            inputs.append(self._prepare_grpc_tensor("end_id", end_id_data))

        return inputs



import time
import urllib.parse
from typing import Any, Dict, Iterator, Optional

import requests
from tenacity import retry, stop_after_delay, wait_fixed
from truss.remote.baseten.api import BasetenApi
from truss.remote.baseten.auth import AuthService
from truss.remote.truss_remote import TrussService
from truss.truss_handle import TrussHandle
from truss.util.errors import RemoteNetworkError

DEFAULT_STREAM_ENCODING = "utf-8"


def _add_model_subdomain(rest_api_url: str, model_subdomain: str) -> str:
    """E.g. `https://api.baseten.co` -> `https://{model_subdomain}.api.baseten.co`"""
    parsed_url = urllib.parse.urlparse(rest_api_url)
    new_netloc = f"{model_subdomain}.{parsed_url.netloc}"
    model_url = parsed_url._replace(netloc=new_netloc)
    return str(urllib.parse.urlunparse(model_url))


class BasetenService(TrussService):
    def __init__(
        self,
        model_id: str,
        model_version_id: str,
        is_draft: bool,
        api_key: str,
        service_url: str,
        api: BasetenApi,
        truss_handle: Optional[TrussHandle] = None,
    ):
        super().__init__(is_draft=is_draft, service_url=service_url)
        self._model_id = model_id
        self._model_version_id = model_version_id
        self._auth_service = AuthService(api_key=api_key)
        self._api = api
        self._truss_handle = truss_handle

    def is_live(self) -> bool:
        raise NotImplementedError

    def is_ready(self) -> bool:
        raise NotImplementedError

    @property
    def model_id(self) -> str:
        return self._model_id

    @property
    def model_version_id(self) -> str:
        return self._model_version_id

    @property
    def invocation_url(self) -> str:
        return f"{self._service_url}/predict"

    def predict(
        self,
        model_request_body: Dict,
    ) -> Any:
        response = self._send_request(
            self.predict_url, "POST", data=model_request_body, stream=True
        )

        if response.headers.get("transfer-encoding") == "chunked":
            # Case of streaming response, the backend does not set an encoding, so
            # manually decode to the contents to utf-8 here.
            def decode_content():
                for chunk in response.iter_content(
                    chunk_size=8192, decode_unicode=True
                ):
                    # Depending on the content-type of the response,
                    # iter_content will either emit a byte stream, or a stream
                    # of strings. Only decode in the bytes case.
                    if isinstance(chunk, bytes):
                        yield chunk.decode(response.encoding or DEFAULT_STREAM_ENCODING)
                    else:
                        yield chunk

            return decode_content()

        parsed_response = response.json()

        if "error" in parsed_response:
            # In the case that the model is in a non-ready state, the response
            # will be a json with an `error` key.
            return parsed_response

        return response.json()

    def authenticate(self) -> dict:
        return self._auth_service.authenticate().header()

    @property
    def logs_url(self) -> str:
        return (
            f"{self._api.remote_url}/models/{self._model_id}/"
            f"logs/{self._model_version_id}"
        )

    @property
    def predict_url(self) -> str:
        """
        Get the URL for the prediction endpoint.
        """
        # E.g. `https://api.baseten.co` -> `https://model-{model_id}.api.baseten.co`
        url = _add_model_subdomain(self._api.rest_api_url, f"model-{self.model_id}")
        if self.is_draft:
            # "https://model-{model_id}.api.baseten.co/development".
            url = f"{url}/development/predict"
        else:
            # "https://model-{model_id}.api.baseten.co/deployment/{deployment_id}".
            url = f"{url}/deployment/{self.model_version_id}/predict"
        return url

    @retry(stop=stop_after_delay(60), wait=wait_fixed(1), reraise=True)
    def _fetch_deployment(self) -> Any:
        return self._api.get_deployment(self._model_id, self._model_version_id)

    def poll_deployment_status(self, sleep_secs: int = 1) -> Iterator[str]:
        """
        Wait for the service to be deployed.
        """
        while True:
            time.sleep(sleep_secs)
            try:
                deployment = self._fetch_deployment()
                yield deployment["status"]
            except requests.exceptions.RequestException:
                raise RemoteNetworkError("Could not reach backend.")



from dataclasses import dataclass
from typing import Dict


@dataclass
class TrussSignature:
    """Truss signature stores information for calculating patches for future
    changes to Truss.

    Currently, it stores hashes of all of the paths in the truss directory excluding the data dir,
    and the truss config contents. Path hashes allow calculating added/updated/removes
    paths in future trusses compared to this. Config contents allow calculating
    config changes, such as add/update/remove of python requirements etc.
    """

    content_hashes_by_path: Dict[str, str]
    config: str

    def to_dict(self) -> dict:
        return {
            "content_hashes_by_path": self.content_hashes_by_path,
            "config": self.config,
        }

    @staticmethod
    def from_dict(d) -> "TrussSignature":
        return TrussSignature(
            content_hashes_by_path=d["content_hashes_by_path"],
            config=d["config"],
        )



# TODO: this file contains too much implementation -> restructure.
import abc
import logging
import os
import pathlib
import traceback
from types import GenericAlias
from typing import (
    Any,
    ClassVar,
    Generic,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
    runtime_checkable,
)

import pydantic
from pydantic import generics
from truss import truss_config
from truss.remote import baseten as baseten_remote
from truss.remote import remote_cli, remote_factory

UserConfigT = TypeVar("UserConfigT", bound=Union[pydantic.BaseModel, None])

BASETEN_API_SECRET_NAME = "baseten_chain_api_key"
SECRET_DUMMY = "***"
TRUSS_CONFIG_CHAINS_KEY = "chains_metadata"

GENERATED_CODE_DIR = ".chains_generated"

# Below arg names must correspond to `definitions.ABCChainlet`.
ENDPOINT_METHOD_NAME = "run_remote"  # Chainlet method name exposed as endpoint.
CONTEXT_ARG_NAME = "context"  # Referring to Chainlets `__init__` signature.
SELF_ARG_NAME = "self"
REMOTE_CONFIG_NAME = "remote_config"


K = TypeVar("K", contravariant=True)
V = TypeVar("V", covariant=True)


@runtime_checkable
class MappingNoIter(Protocol[K, V]):
    def __getitem__(self, key: K) -> V:
        ...

    def __len__(self) -> int:
        ...

    def __contains__(self, key: K) -> bool:
        ...


class SafeModel(pydantic.BaseModel):
    """Pydantic base model with reasonable config."""

    class Config:
        arbitrary_types_allowed = False
        validate_all = True
        validate_assignment = True
        extra = "forbid"


class SafeModelNonSerializable(pydantic.BaseModel):
    """Pydantic base model with reasonable config - allowing arbitrary types."""

    class Config:
        arbitrary_types_allowed = True
        validate_all = True
        validate_assignment = True
        extra = "forbid"


class ChainsUsageError(TypeError):
    """Raised when user-defined Chainlets do not adhere to API constraints."""


class MissingDependencyError(TypeError):
    """Raised when a needed resource could not be found or is not defined."""


class ChainsRuntimeError(Exception):
    """Raised when components are not used the expected way at runtime."""


class AbsPath:
    _abs_file_path: str
    _creating_module: str
    _original_path: str

    def __init__(
        self, abs_file_path: str, creating_module: str, original_path: str
    ) -> None:
        self._abs_file_path = abs_file_path
        self._creating_module = creating_module
        self._original_path = original_path

    def _raise_if_not_exists(self, abs_path: str) -> None:
        if not os.path.isfile(abs_path):
            raise MissingDependencyError(
                f"With the file path `{self._original_path}` an absolute path relative "
                f"to the calling module `{self._creating_module}` was created, "
                f"resulting `{self._abs_file_path}` - but no file was found."
            )

    @property
    def abs_path(self) -> str:
        if self._abs_file_path != self._original_path:
            logging.info(
                f"Using abs path `{self._abs_file_path}` for path specified as "
                f"`{self._original_path}` (in `{self._creating_module}`)."
            )
        abs_path = self._abs_file_path
        self._raise_if_not_exists(abs_path)
        return abs_path


class DockerImage(SafeModelNonSerializable):
    # TODO: this is not stable yet and might change or refer back to truss.
    base_image: str = "python:3.11-slim"
    pip_requirements_file: Optional[AbsPath] = None
    pip_requirements: list[str] = []
    apt_requirements: list[str] = []
    data_dir: Optional[AbsPath] = None


class ComputeSpec(pydantic.BaseModel):
    # TODO: this is not stable yet and might change or refer back to truss.
    cpu_count: int = 1
    predict_concurrency: int = 1
    memory: str = "2Gi"
    accelerator: truss_config.AcceleratorSpec = truss_config.AcceleratorSpec()


CpuCountT = Literal["cpu_count"]
CPU_COUNT: CpuCountT = "cpu_count"


class Compute:
    """Builder to create ComputeSpec."""

    # This extra layer around `ComputeSpec` is needed to parse the accelerator options.

    _spec: ComputeSpec

    def __init__(
        self,
        cpu_count: int = 1,
        memory: str = "2Gi",
        gpu: Union[str, truss_config.Accelerator, None] = None,
        gpu_count: int = 1,
        predict_concurrency: Union[int, CpuCountT] = 1,
    ) -> None:
        accelerator = truss_config.AcceleratorSpec()
        if gpu:
            accelerator.accelerator = truss_config.Accelerator(gpu)
            accelerator.count = gpu_count
            accelerator = truss_config.AcceleratorSpec(
                accelerator=truss_config.Accelerator(gpu), count=gpu_count
            )
        if predict_concurrency == CPU_COUNT:
            predict_concurrency_int = cpu_count
        else:
            assert isinstance(predict_concurrency, int)
            predict_concurrency_int = predict_concurrency

        self._spec = ComputeSpec(
            cpu_count=cpu_count,
            memory=memory,
            accelerator=accelerator,
            predict_concurrency=predict_concurrency_int,
        )

    def get_spec(self) -> ComputeSpec:
        return self._spec.copy(deep=True)


class AssetSpec(SafeModel):
    # TODO: this is not stable yet and might change or refer back to truss.
    secrets: dict[str, str] = {}
    cached: list[Any] = []


class Assets:
    """Builder to create asset spec."""

    # This extra layer around `ComputeSpec` is needed to add secret_keys.
    _spec: AssetSpec

    def __init__(
        self,
        cached: Iterable[Any] = (),
        secret_keys: Iterable[str] = (),
    ) -> None:
        self._spec = AssetSpec(
            cached=list(cached), secrets={k: SECRET_DUMMY for k in secret_keys}
        )

    def get_spec(self) -> AssetSpec:
        return self._spec.copy(deep=True)


class RemoteConfig(SafeModelNonSerializable):
    """Bundles config values needed to deploy a Chainlet."""

    docker_image: DockerImage = DockerImage()
    compute: Compute = Compute()
    assets: Assets = Assets()
    name: Optional[str] = None

    def get_compute_spec(self) -> ComputeSpec:
        return self.compute.get_spec()

    def get_asset_spec(self) -> AssetSpec:
        return self.assets.get_spec()


class RPCOptions(SafeModel):
    timeout_sec: int = 600
    retries: int = 1


class ServiceDescriptor(SafeModel):
    name: str
    predict_url: str
    options: RPCOptions


class DeploymentContext(generics.GenericModel, Generic[UserConfigT]):
    """Bundles config values and resources needed to instantiate Chainlets."""

    class Config:
        arbitrary_types_allowed = True
        validate_all = True
        validate_assignment = True
        extra = "forbid"

    data_dir: Optional[pathlib.Path] = None
    user_config: UserConfigT
    chainlet_to_service: Mapping[str, ServiceDescriptor] = {}
    secrets: MappingNoIter[str, str]

    def get_service_descriptor(self, chainlet_name: str) -> ServiceDescriptor:
        if chainlet_name not in self.chainlet_to_service:
            raise MissingDependencyError(f"{chainlet_name}")
        return self.chainlet_to_service[chainlet_name]

    def get_baseten_api_key(self) -> str:
        if self.secrets is None:
            raise ChainsRuntimeError(
                f"Secrets not set in `{self.__class__.__name__}` object."
            )
        error_msg = (
            "For using chains, it is required to setup a an API key with name "
            f"`{BASETEN_API_SECRET_NAME}` on baseten to allow chain Chainlet to "
            "call other Chainlets. For local execution, secrets can be provided "
            "to `run_local`."
        )
        if BASETEN_API_SECRET_NAME not in self.secrets:
            raise MissingDependencyError(error_msg)

        api_key = self.secrets[BASETEN_API_SECRET_NAME]
        if api_key == SECRET_DUMMY:
            raise MissingDependencyError(
                f"{error_msg}. Retrieved dummy value of `{api_key}`."
            )
        return api_key


class TrussMetadata(generics.GenericModel, Generic[UserConfigT]):
    """Plugin for the truss config (in config["model_metadata"]["chains_metadata"])."""

    user_config: UserConfigT
    chainlet_to_service: Mapping[str, ServiceDescriptor] = {}


class ABCChainlet(abc.ABC):
    remote_config: ClassVar[RemoteConfig] = RemoteConfig(docker_image=DockerImage())
    default_user_config: ClassVar[Optional[pydantic.BaseModel]] = None
    _init_is_patched: ClassVar[bool] = False

    @classmethod
    def has_custom_init(cls) -> bool:
        return cls.__init__ is not object.__init__

    # Cannot add this abstract method to API, because we want to allow arbitrary
    # arg/kwarg names and specifying any function signature here would give type errors
    # @abc.abstractmethod
    # def run_remote(self, *args, **kwargs) -> Any:
    #     ...


class TypeDescriptor(SafeModelNonSerializable):
    """For describing I/O types of Chainlets."""

    raw: Any  # The raw type annotation object (could be a type or GenericAlias).

    @property
    def is_pydantic(self) -> bool:
        return (
            isinstance(self.raw, type)
            and not isinstance(self.raw, GenericAlias)
            and issubclass(self.raw, pydantic.BaseModel)
        )


class EndpointAPIDescriptor(SafeModelNonSerializable):
    name: str = ENDPOINT_METHOD_NAME
    input_names_and_types: list[tuple[str, TypeDescriptor]]
    output_types: list[TypeDescriptor]
    is_async: bool
    is_generator: bool


class DependencyDescriptor(SafeModelNonSerializable):
    chainlet_cls: Type[ABCChainlet]
    options: RPCOptions

    @property
    def name(self) -> str:
        return self.chainlet_cls.__name__


class ChainletAPIDescriptor(SafeModelNonSerializable):
    chainlet_cls: Type[ABCChainlet]
    src_path: str
    has_context: bool
    dependencies: Mapping[str, DependencyDescriptor]
    endpoint: EndpointAPIDescriptor
    user_config_type: TypeDescriptor

    def __hash__(self) -> int:
        return hash(self.chainlet_cls)

    @property
    def name(self) -> str:
        return self.chainlet_cls.__name__


class StackFrame(SafeModel):
    filename: str
    lineno: Optional[int]
    name: str
    line: Optional[str]

    @classmethod
    def from_frame_summary(cls, frame: traceback.FrameSummary):
        return cls(
            filename=frame.filename,
            lineno=frame.lineno,
            name=frame.name,
            line=frame.line,
        )

    def to_frame_summary(self) -> traceback.FrameSummary:
        return traceback.FrameSummary(
            filename=self.filename, lineno=self.lineno, name=self.name, line=self.line
        )


class RemoteErrorDetail(SafeModel):
    remote_name: str
    exception_cls_name: str
    exception_module_name: Optional[str]
    exception_message: str
    user_stack_trace: list[StackFrame]

    def to_stack_summary(self) -> traceback.StackSummary:
        return traceback.StackSummary.from_list(
            frame.to_frame_summary() for frame in self.user_stack_trace
        )

    def format(self) -> str:
        stack = "".join(traceback.format_list(self.to_stack_summary()))
        exc_info = (
            f"\n(Exception class defined in `{self.exception_module_name}`.)"
            if self.exception_module_name
            else ""
        )
        error = (
            f"{RemoteErrorDetail.__name__} in `{self.remote_name}`\n"
            f"Traceback (most recent call last):\n"
            f"{stack}{self.exception_cls_name}: {self.exception_message}{exc_info}"
        )
        return error


class GenericRemoteException(Exception):
    ...


########################################################################################


class DeploymentOptions(SafeModelNonSerializable):
    chain_name: str
    only_generate_trusses: bool = False


class DeploymentOptionsBaseten(DeploymentOptions):
    remote_provider: baseten_remote.BasetenRemote
    publish: bool
    promote: bool

    @classmethod
    def create(
        cls,
        chain_name: str,
        publish: bool,
        promote: bool,
        only_generate_trusses: bool,
        remote: Optional[str] = None,
    ) -> "DeploymentOptionsBaseten":
        if not remote:
            remote = remote_cli.inquire_remote_name(
                remote_factory.RemoteFactory.get_available_config_names()
            )
        remote_provider = cast(
            baseten_remote.BasetenRemote,
            remote_factory.RemoteFactory.create(remote=remote),
        )
        return DeploymentOptionsBaseten(
            remote_provider=remote_provider,
            chain_name=chain_name,
            publish=publish,
            promote=promote,
            only_generate_trusses=only_generate_trusses,
        )


class DeploymentOptionsLocalDocker(DeploymentOptions):
    # Local docker-to-docker requests don't need auth, but we need to set a
    # value different from `SECRET_DUMMY` to not trigger the check that the secret
    # is unset. Additionally, if local docker containers make calls to models deployed
    # on baseten, a real API key must be provided (i.e. the default must be overridden).
    baseten_chain_api_key: str = "docker_dummy_key"



from http import HTTPStatus
from typing import Optional

from fastapi.responses import JSONResponse


class ModelMissingError(Exception):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return self.path


class InferenceError(RuntimeError):
    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.reason


class InvalidInput(ValueError):
    """
    Exception class indicating invalid input arguments.
    HTTP Servers should return HTTP_400 (Bad Request).
    """

    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return self.reason


class ModelNotFound(Exception):
    """
    Exception class indicating requested model does not exist.
    HTTP Servers should return HTTP_404 (Not Found).
    """

    def __init__(self, model_name=None):
        self.reason = f"Model with name {model_name} does not exist."

    def __str__(self):
        return self.reason


class ModelNotReady(RuntimeError):
    def __init__(self, model_name: str, detail: Optional[str] = None):
        self.model_name = model_name
        self.error_msg = f"Model with name {self.model_name} is not ready."
        if detail:
            self.error_msg = self.error_msg + " " + detail

    def __str__(self):
        return self.error_msg


async def exception_handler(_, exc):
    return JSONResponse(
        status_code=HTTPStatus.INTERNAL_SERVER_ERROR, content={"error": str(exc)}
    )


async def invalid_input_handler(_, exc):
    return JSONResponse(status_code=HTTPStatus.BAD_REQUEST, content={"error": str(exc)})


async def inference_error_handler(_, exc):
    return JSONResponse(
        status_code=HTTPStatus.INTERNAL_SERVER_ERROR, content={"error": str(exc)}
    )


async def generic_exception_handler(_, exc):
    return JSONResponse(
        status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
        content={"error": f"{type(exc).__name__} : {str(exc)}"},
    )


async def model_not_found_handler(_, exc):
    return JSONResponse(status_code=HTTPStatus.NOT_FOUND, content={"error": str(exc)})


async def model_not_ready_handler(_, exc):
    return JSONResponse(
        status_code=HTTPStatus.SERVICE_UNAVAILABLE, content={"error": str(exc)}
    )


async def not_implemented_error_handler(_, exc):
    return JSONResponse(
        status_code=HTTPStatus.NOT_IMPLEMENTED, content={"error": str(exc)}
    )


async def http_exception_handler(_, exc):
    return JSONResponse(status_code=exc.status_code, content={"error": exc.detail})



import logging
from pathlib import Path
from typing import Dict, List, Optional, Set

import yaml
from truss.constants import CONFIG_FILE
from truss.patch.hash import file_content_hash_str
from truss.patch.types import TrussSignature
from truss.templates.control.control.helpers.truss_patch.requirement_name_identifier import (
    reqs_by_name,
)
from truss.templates.control.control.helpers.truss_patch.system_packages import (
    system_packages_set,
)
from truss.templates.control.control.helpers.types import (
    Action,
    ConfigPatch,
    DataPatch,
    EnvVarPatch,
    ExternalDataPatch,
    ModelCodePatch,
    PackagePatch,
    Patch,
    PatchType,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.truss_config import ExternalData, TrussConfig
from truss.truss_spec import TrussSpec
from truss.util.path import get_ignored_relative_paths

logger: logging.Logger = logging.getLogger(__name__)
PYCACHE_IGNORE_PATTERNS = [
    "**/__pycache__/**/*",
    "**/__pycache__/**",
]
UNPATCHABLE_CONFIG_KEYS = [
    "live_reload",
    "python_version",
    "resources",
]


def calc_truss_patch(
    truss_dir: Path,
    previous_truss_signature: TrussSignature,
    ignore_patterns: Optional[List[str]] = None,
) -> Optional[List[Patch]]:
    """
    Calculate patch for a truss from a previous state.

    Returns: None if patch cannot be calculated, otherwise a list of patches.
        Note that the none return value is pretty important, patch coverage is
        limited and this usually indicates that the identified change cannot be
        expressed with currently supported patches.

        Only standard and relevant truss paths are scanned for changes, rest
        ignored. E.g. at the root level, only changes to config.yaml are
        checked, any other changes are ignored.
    """

    def _relative_to_root(path: Path) -> str:
        return str(path.relative_to(truss_dir))

    if ignore_patterns is None:
        ignore_patterns = PYCACHE_IGNORE_PATTERNS

    changed_paths = _calc_changed_paths(
        truss_dir,
        previous_truss_signature.content_hashes_by_path,
        ignore_patterns,
    )

    truss_spec = TrussSpec(truss_dir)
    model_module_path = _relative_to_root(truss_spec.model_module_dir)
    data_dir_path = _relative_to_root(truss_spec.data_dir)
    bundled_packages_path = _relative_to_root(truss_spec.bundled_packages_dir)

    def _under_unsupported_patch_dir(path: str) -> bool:
        """
        Checks if the given path is under one of the directories that don't
        support patching. Note that if path `is` one of those directories that's
        ok, because those empty directories can be ignored from patching point
        of view.
        """
        return _strictly_under(path, [data_dir_path])

    patches = []
    for path in changed_paths["removed"]:
        if _strictly_under(path, [model_module_path]):
            logger.info(f"Created patch to remove model code file: {path}")
            patches.append(
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        action=Action.REMOVE,
                        path=_relative_to(path, model_module_path),
                    ),
                )
            )
        elif path == CONFIG_FILE:
            # Don't support removal of config file
            logger.info(f"Patching not supported for removing {path}")
            return None
        elif _strictly_under(path, [bundled_packages_path]):
            logger.info(f"Created patch to remove package file: {path}")
            patches.append(
                Patch(
                    type=PatchType.PACKAGE,
                    body=PackagePatch(
                        action=Action.REMOVE,
                        path=_relative_to(path, bundled_packages_path),
                    ),
                )
            )
        elif _under_unsupported_patch_dir(path):
            logger.warn(f"Patching not supported for removing {path}")
            return None

    for path in changed_paths["added"] + changed_paths["updated"]:
        action = Action.ADD if path in changed_paths["added"] else Action.UPDATE
        if _strictly_under(path, [model_module_path]):
            full_path = truss_dir / path

            # TODO(pankaj) Add support for empty directories, skip them for now.
            if not full_path.is_file():
                continue
            logger.info(
                f"Created patch to {action.value.lower()} model code file: {path}"
            )
            patches.append(
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        action=action,
                        path=_relative_to(path, model_module_path),
                        content=full_path.read_text(),
                    ),
                )
            )
        elif path == CONFIG_FILE:
            new_config = TrussConfig.from_yaml(truss_dir / CONFIG_FILE)
            prev_config = TrussConfig.from_dict(
                yaml.safe_load(previous_truss_signature.config)
            )
            config_patches = calc_config_patches(prev_config, new_config)
            if config_patches:
                logger.info(f"Created patch to {action.value.lower()} config")
            patches.extend(config_patches)
        elif _strictly_under(path, [bundled_packages_path]):
            full_path = truss_dir / path
            if not full_path.is_file():
                continue
            logger.info(f"Created patch to {action.value.lower()} package file: {path}")
            patches.append(
                Patch(
                    type=PatchType.PACKAGE,
                    body=PackagePatch(
                        action=action,
                        path=_relative_to(path, bundled_packages_path),
                        content=full_path.read_text(),
                    ),
                )
            )
        elif _under_unsupported_patch_dir(path):
            logger.info(f"Patching not supported for updating {path}")
            return None
    return patches


def _calc_changed_paths(
    root: Path,
    previous_root_path_content_hashes: Dict[str, str],
    ignore_patterns: Optional[List[str]],
) -> Dict[str, List[str]]:
    """
    TODO(pankaj) add support for directory creation in patch
    """
    root_relative_new_paths = set(
        (str(path.relative_to(root)) for path in root.glob("**/*"))
    )
    unignored_new_paths = calc_unignored_paths(root_relative_new_paths, ignore_patterns)
    previous_root_relative_paths = set(previous_root_path_content_hashes.keys())
    unignored_prev_paths = calc_unignored_paths(
        previous_root_relative_paths, ignore_patterns
    )

    added_paths = unignored_new_paths - unignored_prev_paths
    removed_paths = unignored_prev_paths - unignored_new_paths

    updated_paths = set()
    common_paths = unignored_new_paths.intersection(unignored_prev_paths)
    for path in common_paths:
        full_path: Path = root / path
        if full_path.is_file():
            content_hash = file_content_hash_str(full_path)
            previous_content_hash = previous_root_path_content_hashes[path]
            if content_hash != previous_content_hash:
                updated_paths.add(path)

    return {
        "added": list(added_paths),
        "updated": list(updated_paths),
        "removed": list(removed_paths),
    }


def calc_unignored_paths(
    root_relative_paths: Set[str],
    ignore_patterns: Optional[List[str]] = None,
) -> Set[str]:
    ignored_paths = set(
        get_ignored_relative_paths(root_relative_paths, ignore_patterns)
    )
    return root_relative_paths - ignored_paths  # type: ignore


def calc_config_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to config.

    Returns None if patch cannot be calculated. Empty list means no relevant
    differences found.
    """
    try:
        config_patches = _calc_general_config_patches(prev_config, new_config)
        python_requirement_patches = _calc_python_requirements_patches(
            prev_config, new_config
        )
        system_package_patches = _calc_system_packages_patches(prev_config, new_config)
        return [*config_patches, *python_requirement_patches, *system_package_patches]
    except Exception as e:
        logger.error(f"Failed to calculate config patch with exception: {e}")
        raise


def _calc_general_config_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to config.yaml
    If a change has been made to the config, at least one ConfigPatch is created
    Additional patches are created for each of the patches that need specific application,
    namely patches to the env variables, external package dirs, and/or external data.

    Empty list means no relevant differences found.
    """

    patches: List[Patch] = []
    for key in UNPATCHABLE_CONFIG_KEYS:
        prev_items = getattr(prev_config, key)
        new_items = getattr(new_config, key)
        if new_items != prev_items:
            logger.warning(f"Patching is not supported for: {key}")

    if prev_config.to_dict() != new_config.to_dict():
        patches.append(_mk_config_patch(Action.UPDATE, new_config.to_dict()))

    env_var_patches = _calc_env_var_patches(prev_config, new_config)
    external_data_patches = _calc_external_data_patches(prev_config, new_config)

    return [*patches, *env_var_patches, *external_data_patches]


def _calc_env_var_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to environment variables.

    Empty list means no relevant differences found.
    """
    patches = []
    prev_items = prev_config.environment_variables
    new_items = new_config.environment_variables
    prev_item_names = set(prev_items.keys())
    new_item_names = set(new_items.keys())
    removed_items = prev_item_names.difference(new_item_names)
    for removed_item in removed_items:
        patches.append(
            _mk_env_var_patch(
                Action.REMOVE,
                {removed_item: prev_items[removed_item]},
            )
        )
    added_items = new_item_names.difference(prev_item_names)
    for added_item in added_items:
        patches.append(
            _mk_env_var_patch(Action.ADD, {added_item: new_items[added_item]})
        )
    for item in new_item_names.intersection(prev_item_names):
        if new_items[item] != prev_items[item]:
            patches.append(_mk_env_var_patch(Action.UPDATE, {item: new_items[item]}))
    return patches


def _calc_external_data_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to external data.

    Empty list means no relevant differences found.
    """
    patches = []
    prev_items = (prev_config.external_data or ExternalData([])).to_list()
    new_items = (new_config.external_data or ExternalData([])).to_list()

    removed_items = [x for x in prev_items if x not in new_items]
    for removed_item in removed_items:
        patches.append(
            _mk_external_data_patch(
                Action.REMOVE,
                removed_item,
            )
        )
    added_items = [x for x in new_items if x not in prev_items]
    for added_item in added_items:
        patches.append(_mk_external_data_patch(Action.ADD, added_item))
    return patches


def _calc_python_requirements_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to python requirements.

    Empty list means no relevant differences found.
    """
    patches = []
    prev_reqs = reqs_by_name(prev_config.requirements)
    prev_req_names = set(prev_reqs.keys())
    new_reqs = reqs_by_name(new_config.requirements)
    new_req_names = set(new_reqs.keys())
    removed_reqs = prev_req_names.difference(new_req_names)
    for removed_req in removed_reqs:
        patches.append(_mk_python_requirement_patch(Action.REMOVE, removed_req))

    added_reqs = new_req_names.difference(prev_req_names)
    for added_req in added_reqs:
        patches.append(_mk_python_requirement_patch(Action.ADD, new_reqs[added_req]))

    for req in new_req_names.intersection(prev_req_names):
        if new_reqs[req] != prev_reqs[req]:
            patches.append(_mk_python_requirement_patch(Action.UPDATE, new_reqs[req]))

    return patches


def _calc_system_packages_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to system packates.

    Empty list means no relevant differences found.
    """
    patches = []
    prev_pkgs = system_packages_set(prev_config.system_packages)
    new_pkgs = system_packages_set(new_config.system_packages)
    removed_pkgs = prev_pkgs.difference(new_pkgs)
    for removed_pkg in removed_pkgs:
        patches.append(_mk_system_package_patch(Action.REMOVE, removed_pkg))

    added_pkgs = new_pkgs.difference(prev_pkgs)
    for added_pkg in added_pkgs:
        patches.append(_mk_system_package_patch(Action.ADD, added_pkg))

    return patches


def _mk_config_patch(action: Action, config: dict) -> Patch:
    return Patch(
        type=PatchType.CONFIG,
        body=ConfigPatch(
            action=action,
            config=config,
        ),
    )


# Support for patching data changes yet to be implemented
def _mk_data_patch(action: Action, item: str, path: str) -> Patch:
    return Patch(
        type=PatchType.DATA,
        body=DataPatch(action=action, content=item, path=path),
    )


def _mk_env_var_patch(action: Action, item: dict) -> Patch:
    return Patch(
        type=PatchType.ENVIRONMENT_VARIABLE,
        body=EnvVarPatch(
            action=action,
            item=item,
        ),
    )


def _mk_external_data_patch(action: Action, item: Dict[str, str]) -> Patch:
    return Patch(
        type=PatchType.EXTERNAL_DATA,
        body=ExternalDataPatch(
            action=action,
            item=item,
        ),
    )


def _mk_python_requirement_patch(action: Action, requirement: str) -> Patch:
    return Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(
            action=action,
            requirement=requirement,
        ),
    )


def _mk_system_package_patch(action: Action, package: str) -> Patch:
    return Patch(
        type=PatchType.SYSTEM_PACKAGE,
        body=SystemPackagePatch(
            action=action,
            package=package,
        ),
    )


def _relative_to(path: str, relative_to_path: str):
    return str(Path(path).relative_to(relative_to_path))


def _strictly_under(path: str, parent_paths: List[str]) -> bool:
    """
    Checks if given path is under one of the given paths, but not the same as
    them. Assumes that parent paths themselves are not under each other.
    """
    for dir_path in parent_paths:
        if path.startswith(dir_path) and not path == dir_path:
            return True
    return False



import inspect
import json
import logging
import os
import sys
import time
from functools import wraps
from pathlib import Path
from typing import Callable, List, Optional, Tuple

import rich
import rich.live
import rich.spinner
import rich.table
import rich_click as click
import truss
from truss.cli.console import console
from truss.cli.create import ask_name
from truss.remote.baseten.core import (
    ACTIVE_STATUS,
    DEPLOYING_STATUSES,
    ModelId,
    ModelIdentifier,
    ModelName,
    ModelVersionId,
)
from truss.remote.baseten.service import BasetenService
from truss.remote.remote_cli import inquire_model_name, inquire_remote_name
from truss.remote.remote_factory import USER_TRUSSRC_PATH, RemoteFactory
from truss.truss_config import Build, ModelServer
from truss.util.errors import RemoteNetworkError
from truss_chains import definitions as chains_def
from truss_chains import deploy as chains_deploy
from truss_chains import framework

logging.basicConfig(level=logging.INFO)

rich.spinner.SPINNERS["deploying"] = {"interval": 500, "frames": ["👾 ", " 👾"]}
rich.spinner.SPINNERS["building"] = {"interval": 500, "frames": ["🛠️ ", " 🛠️"]}
rich.spinner.SPINNERS["loading"] = {"interval": 500, "frames": ["⏱️ ", " ⏱️"]}
rich.spinner.SPINNERS["active"] = {"interval": 500, "frames": ["💚 ", " 💚"]}
rich.spinner.SPINNERS["failed"] = {"interval": 500, "frames": ["😤 ", " 😤"]}

click.rich_click.COMMAND_GROUPS = {
    "truss": [
        {
            "name": "Main usage",
            "commands": ["init", "push", "watch", "predict"],
            "table_styles": {  # type: ignore
                "row_styles": ["green"],
            },
        },
        {
            "name": "Advanced Usage",
            "commands": ["image", "container", "cleanup"],
            "table_styles": {  # type: ignore
                "row_styles": ["yellow"],
            },
        },
        {
            "name": "Chains",
            "commands": ["chains"],
            "table_styles": {  # type: ignore
                "row_styles": ["red"],
            },
        },
    ]
}


def echo_output(f: Callable[..., object]):
    @wraps(f)
    def wrapper(*args, **kwargs):
        click.echo(f(*args, **kwargs))

    return wrapper


def error_handling(f: Callable[..., object]):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            f(*args, **kwargs)
        except click.UsageError as e:
            raise e  # You can re-raise the exception or handle it different
        except Exception as e:
            click.secho(f"ERROR ({type(e)}: {e}", fg="red")

    return wrapper


def print_help() -> None:
    ctx = click.get_current_context()
    click.echo(ctx.get_help())


@click.group(name="truss", invoke_without_command=True)  # type: ignore
@click.pass_context
@click.version_option(truss.version())
def truss_cli(ctx) -> None:
    """truss: The simplest way to serve models in production"""
    if not ctx.invoked_subcommand:
        click.echo(ctx.get_help())


@click.group()
def container():
    """Subcommands for truss container"""


@click.group()
def image():
    """Subcommands for truss image"""


@click.group()
def chains():
    """Subcommands for truss chains"""


@truss_cli.command()
@click.argument("target_directory", required=True)
@click.option(
    "-b",
    "--backend",
    show_default=True,
    default=ModelServer.TrussServer.value,
    type=click.Choice([server.value for server in ModelServer]),
)
@error_handling
def init(target_directory, backend) -> None:
    """Create a new truss.

    TARGET_DIRECTORY: A Truss is created in this directory
    """
    if os.path.isdir(target_directory):
        raise click.ClickException(
            f"Error: Directory `{target_directory}` already exists "
            "and cannot be overwritten."
        )
    tr_path = Path(target_directory)
    build_config = Build(model_server=ModelServer[backend])
    model_name = ask_name()
    truss.init(
        target_directory=target_directory,
        build_config=build_config,
        model_name=model_name,
    )
    click.echo(f"Truss {model_name} was created in {tr_path.absolute()}")


@image.command()  # type: ignore
@click.argument("build_dir")
@click.argument("target_directory", required=False)
@error_handling
def build_context(build_dir, target_directory: str) -> None:
    """
    Create a docker build context for a Truss.

    BUILD_DIR: Folder where image context is built for Truss

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    tr.docker_build_setup(build_dir=Path(build_dir))


@image.command()  # type: ignore
@click.argument("target_directory", required=False)
@click.argument("build_dir", required=False)
@error_handling
@click.option("--tag", help="Docker image tag")
def build(target_directory: str, build_dir: Path, tag) -> None:
    """
    Builds the docker image for a Truss.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    BUILD_DIR: Image context. If none, a temp directory is created.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    if build_dir:
        build_dir = Path(build_dir)
    tr.build_serving_docker_image(build_dir=build_dir, tag=tag)


@image.command()  # type: ignore
@click.argument("target_directory", required=False)
@click.argument("build_dir", required=False)
@click.option("--tag", help="Docker build image tag")
@click.option("--port", type=int, default=8080, help="Local port used to run image")
@click.option(
    "--attach", is_flag=True, default=False, help="Flag for attaching the process"
)
@error_handling
def run(target_directory: str, build_dir: Path, tag, port, attach) -> None:
    """
    Runs the docker image for a Truss.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    BUILD_DIR: Image context. If none, a temp directory is created.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    if build_dir:
        build_dir = Path(build_dir)
    urls = tr.get_urls_from_truss()
    if urls:
        click.confirm(
            f"Container already exists at {urls}. Are you sure you want to continue?"
        )
    tr.docker_run(build_dir=build_dir, tag=tag, local_port=port, detach=not attach)


@truss_cli.command()
@click.argument("target_directory", required=False, default=os.getcwd())
@click.option(
    "--remote",
    type=str,
    required=False,
    help="Name of the remote in .trussrc to patch changes to",
)
@error_handling
def watch(
    target_directory: str,
    remote: str,
) -> None:
    """
    Seamless remote development with truss

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    # TODO: ensure that provider support draft
    if not remote:
        remote = inquire_remote_name(RemoteFactory.get_available_config_names())

    remote_provider = RemoteFactory.create(remote=remote)

    tr = _get_truss_from_directory(target_directory=target_directory)
    model_name = tr.spec.config.model_name
    if not model_name:
        rich.print(
            "🧐 NoneType model_name provided in config.yaml. "
            "Please check that you have the correct model name in your config file."
        )
        sys.exit(1)

    service = remote_provider.get_service(model_identifier=ModelName(model_name))
    rich.print(f"🪵  View logs for your deployment at {service.logs_url}")
    remote_provider.sync_truss_to_dev_version_by_name(model_name, target_directory)


def _create_chains_table(service) -> Tuple[rich.table.Table, List[str]]:
    """Creates a status table e.g.

                                                     Deployments
    ╭──────────────────────┬──────────────────────┬────────────────────────────╮
    │ Status               │ Chainlet             │                   Logs URL │
    ├──────────────────────┼──────────────────────┼────────────────────────────┤
    │  👾 DEPLOYING        │ SplitText            │ https://app.baseten.co/... │
    │  👾 DEPLOYING        │ GenerateData         │ https://app.baseten.co/... │
    │  👾 DEPLOYING        │ MistralLLM           │ https://app.baseten.co/... │
    │  👾 DEPLOYING        │ TextToNum            │ https://app.baseten.co/... │
    │  👾 DEPLOYING        │ Chain                │ https://app.baseten.co/... │
    ╰──────────────────────┴──────────────────────┴────────────────────────────╯
    """
    table = rich.table.Table(
        show_header=True,
        header_style="bold yellow",
        title=f"⛓️   {service.name} - Chainlets  ⛓️",
        box=rich.table.box.ROUNDED,
        border_style="blue",
    )
    table.add_column("Status", style="dim", min_width=20)
    table.add_column("Name", min_width=20)
    table.add_column("Logs URL")
    statuses = []
    # After reversing, the first one is the entrypoint (per order in service).
    for i, (name, status, logs_url) in enumerate(reversed(service.get_info())):
        if status == ACTIVE_STATUS:
            spinner_name = "active"
        elif status in DEPLOYING_STATUSES:
            if status == "BUILDING":
                spinner_name = "building"
            elif status == "LOADING":
                spinner_name = "loading"
            else:
                spinner_name = "deploying"
        else:
            spinner_name = "failed"
        spinner = rich.spinner.Spinner(spinner_name, text=status)
        if i == 0:
            display_name = f"{name} (entrypoint)"
        else:
            display_name = f"{name} (dep)"

        table.add_row(spinner, display_name, logs_url)
        if i == 0:  # Add section divider after entrypoint.
            table.add_section()
        statuses.append(status)
    return table, statuses


@chains.command()  # type: ignore
@click.argument("source", type=Path, required=True)
@click.argument("entrypoint", type=str, required=False)
@click.option(
    "--name",
    type=str,
    required=False,
    help="Name of the chain to be deployed, if not given, the entrypoint name is used.",
)
@click.option(
    "--publish",
    type=bool,
    default=False,
    is_flag=True,
    help="Create chainlets as published deployments.",
)
@click.option(
    "--promote",
    type=bool,
    default=False,
    is_flag=True,
    help="Replace production chainlets with newly deployed chainlets.",
)
@click.option(
    "--wait",
    type=bool,
    default=True,
    is_flag=True,
    help="Wait until all chainlets are ready (or deployment failed).",
)
@click.option(
    "--dryrun",
    type=bool,
    default=False,
    is_flag=True,
    help="Produces only generated files, but doesn't deploy anything.",
)
@error_handling
def deploy(
    source: Path,
    entrypoint: Optional[str],
    name: Optional[str],
    publish: bool,
    promote: bool,
    wait: bool,
    dryrun: bool,
) -> None:
    """
    Deploys a chain remotely.

    SOURCE: Path to a python file that contains the entrypoint chainlet.

    ENTRYPOINT: Class name of the entrypoint chainlet in source file. May be omitted
    if a chainlet definition in SOURCE is tagged with `@chains.entrypoint`.
    """
    with framework.import_target(source, entrypoint) as entrypoint_cls:
        chain_name = name or entrypoint_cls.__name__
        options = chains_def.DeploymentOptionsBaseten.create(
            chain_name=chain_name,
            promote=promote,
            publish=publish,
            only_generate_trusses=dryrun,
        )
        service = chains_deploy.deploy_remotely(entrypoint_cls, options)

    console.print("\n")
    if dryrun:
        return

    run_help_msg = (
        f"curl -X POST '{service.run_url}' \\\n"
        '    -H "Authorization: Api-Key $BASETEN_API_KEY" \\\n'
        "    -d '<JSON_INPUT>'"
    )

    table, statuses = _create_chains_table(service)
    if wait:
        num_services = len(statuses)
        success = False
        num_failed = 0
        with rich.live.Live(table, console=console, refresh_per_second=4) as live:
            while True:
                table, statuses = _create_chains_table(service)
                live.update(table)
                num_active = sum(s == ACTIVE_STATUS for s in statuses)
                num_deploying = sum(s in DEPLOYING_STATUSES for s in statuses)
                if num_active == num_services:
                    success = True
                    break
                elif num_failed := num_services - num_active - num_deploying:
                    break
        # Print must be outside `Live` context.
        if success:
            console.print("Deployment succeeded.", style="bold green")
            console.print(f"You can run the chain with:\n{run_help_msg}")
        else:
            console.print(f"Deployment failed ({num_failed} failures).", style="red")
    else:
        console.print(table)
        console.print(
            "Once all chainlets are deployed, "
            f"you can run the chain with:\n\n{run_help_msg}"
        )


def _extract_and_validate_model_identifier(
    target_directory: str,
    model_id: Optional[str],
    model_version_id: Optional[str],
    published: Optional[bool],
) -> ModelIdentifier:
    if published and (model_id or model_version_id):
        raise click.UsageError(
            "Cannot use --published with --model or --model-deployment."
        )

    model_identifier: ModelIdentifier
    if model_version_id:
        model_identifier = ModelVersionId(model_version_id)
    elif model_id:
        model_identifier = ModelId(model_id)
    else:
        tr = _get_truss_from_directory(target_directory=target_directory)
        model_name = tr.spec.config.model_name
        if not model_name:
            raise click.UsageError("Truss config is missing a model name.")
        model_identifier = ModelName(model_name)
    return model_identifier


def _extract_request_data(data: Optional[str], file: Optional[Path]):
    try:
        if data is not None:
            return json.loads(data)
        if file is not None:
            return json.loads(Path(file).read_text())
    except json.JSONDecodeError:
        raise click.UsageError("Request data must be valid json.")

    raise click.UsageError(
        "You must provide exactly one of '--data (-d)' or '--file (-f)' options."
    )


@truss_cli.command()
@click.option("--target-directory", required=False, help="Directory of truss")
@click.option(
    "--remote",
    type=str,
    required=False,
    help="Name of the remote in .trussrc to push to",
)
@click.option(
    "-d",
    "--data",
    type=str,
    required=False,
    help="String formatted as json that represents request",
)
@click.option(
    "-f",
    "--file",
    type=click.Path(exists=True),
    help="Path to json file containing the request",
)
@click.option(
    "--published",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help="Call the published model deployment.",
)
@click.option(
    "--model-version",
    type=str,
    required=False,
    help=(
        "[DEPRECATED] Use --model-deployment instead, this will be  "
        "removed in future release. ID of model deployment"
    ),
)
@click.option(
    "--model-deployment",
    type=str,
    required=False,
    help="ID of model deployment to call",
)
@click.option(
    "--model",
    type=str,
    required=False,
    help="ID of model to call",
)
@echo_output
def predict(
    target_directory: str,
    remote: str,
    data: Optional[str],
    file: Optional[Path],
    published: Optional[bool],
    model_version: Optional[str],
    model_deployment: Optional[str],
    model: Optional[str],
):
    """
    Calls the packaged model

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    REQUEST: String formatted as json that represents request

    REQUEST_FILE: Path to json file containing the request
    """
    if not remote:
        remote = inquire_remote_name(RemoteFactory.get_available_config_names())

    remote_provider = RemoteFactory.create(remote=remote)

    if model_version:
        console.print(
            "[DEPRECATED] --model-version is deprecated, "
            "use --model-deployment instead.",
            style="yellow",
        )
        model_deployment = model_version

    model_identifier = _extract_and_validate_model_identifier(
        target_directory,
        model_id=model,
        model_version_id=model_deployment,
        published=published,
    )

    request_data = _extract_request_data(data=data, file=file)

    service = remote_provider.get_service(
        model_identifier=model_identifier, published=published
    )

    # Log deployment ID for Baseten models.
    if isinstance(service, BasetenService):
        rich.print(
            f"Calling predict on {'[cyan]development[/cyan] ' if service.is_draft else ''}"
            f"deployment ID {service.model_version_id}..."
        )

    result = service.predict(request_data)
    if inspect.isgenerator(result):
        for chunk in result:
            click.echo(chunk, nl=False)
        return
    rich.print_json(data=result)


@truss_cli.command()
@click.argument("target_directory", required=False, default=os.getcwd())
@click.option(
    "--remote",
    type=str,
    required=False,
    help="Name of the remote in .trussrc to push to",
)
@click.option("--model-name", type=str, required=False, help="Name of the model")
@click.option(
    "--publish",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help=(
        "Push the truss as a published deployment. If no production "
        "deployment exists, promote the truss to production "
        "after deploy completes."
    ),
)
@click.option(
    "--promote",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help=(
        "Push the truss as a published deployment. Even if a production "
        "deployment exists, promote the truss to production "
        "after deploy completes."
    ),
)
@click.option(
    "--preserve-previous-production-deployment",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help=(
        "Preserve the previous production deployment's autoscaling setting. When "
        "not specified, the previous production deployment will be updated to allow "
        "it to scale to zero. Can only be use in combination with --promote option."
    ),
)
@click.option(
    "--trusted",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help="Trust truss with hosted secrets.",
)
@click.option(
    "--deployment-name",
    type=str,
    required=False,
    help=(
        "Name of the deployment created by the push. Can only be "
        "used in combination with `--publish` or `--promote`."
    ),
)
@click.option(
    "--wait/--no-wait",
    type=bool,
    is_flag=True,
    required=False,
    default=False,
    help="Wait for the deployment to complete before returning.",
)
@click.option(
    "--timeout-seconds",
    type=int,
    required=False,
    help=(
        "Maximum time to wait for deployment to complete in seconds. Without "
        "specifying, the command will not complete until the deployment is complete."
    ),
)
@error_handling
def push(
    target_directory: str,
    remote: str,
    model_name: str,
    publish: bool = False,
    trusted: bool = False,
    promote: bool = False,
    preserve_previous_production_deployment: bool = False,
    deployment_name: Optional[str] = None,
    wait: bool = False,
    timeout_seconds: Optional[int] = None,
) -> None:
    """
    Pushes a truss to a TrussRemote.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.

    """
    if not remote:
        remote = inquire_remote_name(RemoteFactory.get_available_config_names())

    remote_provider = RemoteFactory.create(remote=remote)

    tr = _get_truss_from_directory(target_directory=target_directory)

    model_name = model_name or tr.spec.config.model_name
    if not model_name:
        model_name = inquire_model_name()

    # Write model name to config if it's not already there
    if model_name != tr.spec.config.model_name:
        tr.spec.config.model_name = model_name
        tr.spec.config.write_to_yaml_file(tr.spec.config_path, verbose=False)

    # Log a warning if using secrets without --trusted.
    # TODO(helen): this could be moved to a separate function that includes more
    #  config checks.
    if tr.spec.config.secrets and not trusted:
        not_trusted_text = (
            "Warning: your Truss has secrets but was not pushed with --trusted. "
            "Please push with --trusted to grant access to secrets."
        )
        console.print(not_trusted_text, style="red")

    # TODO(Abu): This needs to be refactored to be more generic
    service = remote_provider.push(
        tr,
        model_name=model_name,
        publish=publish,
        trusted=trusted,
        promote=promote,
        preserve_previous_prod_deployment=preserve_previous_production_deployment,
        deployment_name=deployment_name,
    )  # type: ignore

    click.echo(f"✨ Model {model_name} was successfully pushed ✨")

    if service.is_draft:
        draft_model_text = """
|---------------------------------------------------------------------------------------|
| Your model is deploying as a development model. Development models allow you to  |
| iterate quickly during the deployment process.                                        |
|                                                                                       |
| When you are ready to publish your deployed model as a new deployment,                |
| pass `--publish` to the `truss push` command. To monitor changes to your model and    |
| rapidly iterate, run the `truss watch` command.                                       |
|                                                                                       |
|---------------------------------------------------------------------------------------|
"""

        click.echo(draft_model_text)

    if promote:
        promotion_text = (
            "Your Truss has been deployed as a production model. After it successfully "
            "deploys, it will become the next production deployment of your model."
        )
        console.print(promotion_text, style="green")

    rich.print(f"🪵  View logs for your deployment at {service.logs_url}")
    if wait:
        start_time = time.time()
        with console.status("[bold green]Deploying...") as status:
            try:
                # Poll for the deployment status until we have reached. Either ACTIVE,
                # or a non-deploying status (in which case the deployment has failed).
                for deployment_status in service.poll_deployment_status():
                    if (
                        timeout_seconds is not None
                        and time.time() - start_time > timeout_seconds
                    ):
                        console.print("Deployment timed out.", style="red")
                        sys.exit(1)

                    status.update(
                        f"[bold green]Deploying...Current Status: {deployment_status}"
                    )

                    if deployment_status == ACTIVE_STATUS:
                        console.print("Deployment succeeded.", style="bold green")
                        return

                    if deployment_status not in DEPLOYING_STATUSES:
                        console.print(
                            f"Deployment failed with status {deployment_status}.",
                            style="red",
                        )
                        sys.exit(1)

            except RemoteNetworkError:
                console.print("Deployment failed: Could not reach remote.", style="red")
                sys.exit(1)


@truss_cli.command()
def configure():
    # Read the original file content
    with open(USER_TRUSSRC_PATH, "r") as f:
        original_content = f.read()

    # Open the editor and get the modified content
    edited_content = click.edit(original_content)

    # If the content was modified, save it
    if edited_content is not None and edited_content != original_content:
        with open(USER_TRUSSRC_PATH, "w") as f:
            f.write(edited_content)
            click.echo(f"Changes saved to {USER_TRUSSRC_PATH}")
    else:
        click.echo("No changes made.")


@container.command()  # type: ignore
@click.argument("target_directory", required=False)
@error_handling
def logs(target_directory) -> None:
    """
    Get logs in a container is running for a truss

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    for log in _get_truss_from_directory(
        target_directory=target_directory
    ).serving_container_logs():
        click.echo(log)


@container.command()  # type: ignore
@click.argument("target_directory", required=False)
def kill(target_directory: str) -> None:
    """
    Kills containers related to truss.

    TARGET_DIRECTORY: A Truss directory. If none, use current directory.
    """
    tr = _get_truss_from_directory(target_directory=target_directory)
    tr.kill_container()


@container.command()  # type: ignore
def kill_all() -> None:
    """Kills all truss containers that are not manually persisted."""
    truss.kill_all()


@truss_cli.command()
@error_handling
def cleanup() -> None:
    """
    Clean up truss data.

    Truss creates temporary directories for various operations
    such as for building docker images. This command clears
    that data to free up disk space.
    """
    truss.build.cleanup()


def _get_truss_from_directory(target_directory: Optional[str] = None):
    """Gets Truss from directory. If none, use the current directory"""
    if target_directory is None:
        target_directory = os.getcwd()
    return truss.load(target_directory)


truss_cli.add_command(container)
truss_cli.add_command(image)
truss_cli.add_command(chains)

if __name__ == "__main__":
    truss_cli()



import os
import random
import signal
import socket
import sys
import tempfile
import time
from contextlib import contextmanager
from dataclasses import dataclass
from multiprocessing import Process
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Callable

import psutil
import pytest
import requests

PATCH_PING_MAX_DELAY_SECS = 3


@dataclass
class ControlServerDetails:
    control_server_process: Process
    control_server_port: int = 10123
    inference_server_port: int = 10124


@pytest.fixture
def control_server(truss_control_container_fs):
    with _configured_control_server(truss_control_container_fs) as server:
        yield server


@pytest.mark.integration
def test_truss_control_server_termination(control_server: ControlServerDetails):
    # Port should have been taken up by the servers
    proc_id = control_server.control_server_process.pid
    assert not _is_port_available(control_server.control_server_port)
    assert not _is_port_available(control_server.inference_server_port)

    os.kill(proc_id, signal.SIGTERM)
    control_server.control_server_process.join(timeout=30)
    assert not control_server.control_server_process.is_alive()
    assert _process_tree_is_dead(proc_id)


@pytest.mark.integration
def test_truss_control_server_predict_delays(control_server: ControlServerDetails):
    # Patch to identity code
    identity_model_code = """
class Model:
    def predict(self, model_input):
        return model_input
"""

    ctrl_url = f"http://localhost:{control_server.control_server_port}"
    _patch(identity_model_code, control_server)

    # run predictions and verify
    num_requests = 100

    def predict(inp):
        time.sleep(random.uniform(0, 0.5))
        resp = requests.post(
            f"{ctrl_url}/v1/models/model:predict",
            json=inp,
        )
        return resp.json()

    with ThreadPool(10) as p:
        inputs = list(range(0, num_requests))
        predictions = p.map(predict, inputs)
        assert predictions == inputs


@pytest.mark.integration
def test_truss_control_server_stream(control_server: ControlServerDetails):
    # Patch to identity code
    stream_model_code = """
class Model:
    def predict(self, model_input):
        def inner():
            for i in range(5):
                yield str(i)
        return inner()
"""

    ctrl_url = f"http://localhost:{control_server.control_server_port}"
    _patch(stream_model_code, control_server)

    resp = requests.post(f"{ctrl_url}/v1/models/model:predict", json={}, stream=True)
    assert resp.headers.get("transfer-encoding") == "chunked"
    assert resp.content == "01234".encode("utf-8")


@pytest.mark.integration
def test_truss_control_server_patch_ping_delays(truss_control_container_fs: Path):
    for _ in range(10):
        with _configured_control_server(
            truss_control_container_fs,
            with_patch_ping_flow=True,
        ) as control_server:
            # Account for patch ping delays
            time.sleep(PATCH_PING_MAX_DELAY_SECS)
            # Port should have been taken up by the servers
            proc_id = control_server.control_server_process.pid
            _assert_with_retry(
                lambda: not _is_port_available(control_server.control_server_port),
                "control server port is still available",
            )
            _assert_with_retry(
                lambda: not _is_port_available(control_server.inference_server_port),
                "inference server port is still available",
            )

            os.kill(proc_id, signal.SIGTERM)
            control_server.control_server_process.join(timeout=30)
            # Control server process tree should be dead
            assert not control_server.control_server_process.is_alive()
            assert _process_tree_is_dead(proc_id)


def _assert_with_retry(
    pred: Callable[[], bool],
    msg: str,
    retry_interval_secs: float = 0.5,
    retry_count: int = 60,
):
    for _ in range(0, retry_count):
        if pred():
            return
        time.sleep(retry_interval_secs)
    assert False, msg


def _is_port_available(port):
    try:
        # Try to bind to the given port
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("localhost", port))
            return True
    except socket.error:
        # Port is already in use
        return False


def _kill_process_tree(pid: int):
    try:
        proc = psutil.Process(pid)
        for child_proc in proc.children(recursive=True):
            child_proc.kill()
        proc.kill()
    except psutil.NoSuchProcess:
        pass


def _process_tree_is_dead(pid: int):
    try:
        proc = psutil.Process(pid)
    except psutil.NoSuchProcess:
        return True

    if proc.is_running():
        return False
    for child_proc in proc.children(recursive=True):
        if child_proc.is_running():
            return False
    return True


@contextmanager
def _configured_control_server(
    truss_control_container_fs: Path,
    with_patch_ping_flow: bool = False,
):
    # Pick random ports to reduce reuse, port release may take time
    # which can interfere with tests
    ctrl_port = random.randint(10000, 11000)
    inf_port = ctrl_port + 1
    patch_ping_server_port = ctrl_port + 2

    def start_truss_server(stdout_capture_file_path):
        if with_patch_ping_flow:
            os.environ[
                "PATCH_PING_URL_TRUSS"
            ] = f"http://localhost:{patch_ping_server_port}"
        sys.stdout = open(stdout_capture_file_path, "w")
        app_path = truss_control_container_fs / "app"
        sys.path.append(str(app_path))
        control_path = truss_control_container_fs / "control" / "control"
        sys.path.append(str(control_path))

        from server import ControlServer

        control_server = ControlServer(
            python_executable_path=sys.executable,
            inf_serv_home=str(app_path),
            control_server_port=ctrl_port,
            inference_server_port=inf_port,
        )
        control_server.run()

    def start_patch_ping_server():
        import json
        import random
        import time
        from http.server import BaseHTTPRequestHandler, HTTPServer

        class Handler(BaseHTTPRequestHandler):
            def do_POST(self):
                time.sleep(random.uniform(0, PATCH_PING_MAX_DELAY_SECS))
                self.send_response(200)
                self.end_headers()
                self.wfile.write(
                    bytes(json.dumps({"is_current": True}), encoding="utf-8")
                )

        httpd = HTTPServer(("localhost", patch_ping_server_port), Handler)
        httpd.serve_forever()

    stdout_capture_file = tempfile.NamedTemporaryFile()
    subproc = Process(target=start_truss_server, args=(stdout_capture_file.name,))
    subproc.start()
    proc_id = subproc.pid
    if with_patch_ping_flow:
        patch_ping_server_proc = Process(target=start_patch_ping_server)
        patch_ping_server_proc.start()
    try:
        time.sleep(2.0)
        # Port should have been taken up by truss server
        yield ControlServerDetails(
            control_server_process=subproc,
            control_server_port=ctrl_port,
            inference_server_port=inf_port,
        )
    finally:
        if with_patch_ping_flow:
            patch_ping_server_proc.kill()
        # Print on purpose for help with debugging, otherwise hard to know what's going on
        print(Path(stdout_capture_file.name).read_text())
        _kill_process_tree(proc_id)


def _patch(model_code: str, control_server: ControlServerDetails):
    ctrl_url = f"http://localhost:{control_server.control_server_port}"
    resp = requests.get(f"{ctrl_url}/control/truss_hash")
    truss_hash = resp.json()["result"]

    resp = requests.post(
        f"{ctrl_url}/control/patch",
        json={
            "hash": "dummy",
            "prev_hash": truss_hash,
            "patches": [
                {
                    "type": "model_code",
                    "body": {
                        "action": "UPDATE",
                        "path": "model.py",
                        "content": model_code,
                    },
                }
            ],
        },
    )
    resp.raise_for_status()
